class RateController:
    """
    Create RateController class for injection rate control.

    The class provides methods for controlling the injection rate in a step rate test.

    Attributes
    --------
    step_dq: float
        The rate step in m^3/s. Sign convention is negative for injection.
    step_dt: float
        The step duration in seconds, assumed constant for all steps, but the control logic can also handle variable step duration.
    n_steps: int
        The number of steps in step rate test.
    min_dq: float
        The minimum rate step (in absolute value) in m^3/s.
    max_dq: float
        The maximum rate step (in absolute value) in m^3/s.
    q_inj: float
        The injection rate at the current step in m^3/s. Sign convention is negative for injection.
    q_inj_previous_step: float
        The injection rate at the previous step in m^3/s. Sign convention is negative for injection.
    step_start_time: float
        The current step start time in seconds, relative to the simulation start time.
    step_end_time: float
        The current step end time in seconds, relative to the simulation start time.
    step_index: int
        The current step index.
    return_to_previous_injection_step: boolean
        If this flag is True, the controller will use the injection rate from the previous step.

    Methods
    --------
    reduce_rate_step():
        Reduce the rate step by a multiplication factor.
    increase_rate_step():
        Increase the rate step by a multiplication factor.
    react_to_deviation():
        React to a deviation from the safe operating envelope.
    update_parameters_for_next_step():
        Update controller parameters for the next injection rate step.
    """

    def __init__(
            self,
            dq,
            dt,
            n_steps,
            min_dq=-400.0 / 3600 / 24,
            max_dq=-5000.0 / 3600 / 24
    ):
        self.step_dq = dq
        self.step_dt = dt
        self.n_steps = n_steps
        self.min_dq = min_dq
        self.max_dq = max_dq
        self.q_inj = dq
        self.q_inj_previous_step = 0
        self.step_start_time = 0
        self.step_end_time = dt
        self.step_index = 1
        self.return_to_previous_injection_step = False

    def reduce_rate_step(self, factor=0.5):
        """
        Reduce the rate step by a multiplication factor.

        This function applies a multiplication factor between 0 and 1 to reduce the current rate step.
        The rate step is limited to a pre-defined minimum value.

        Parameters
        --------
        factor: float
            Multiplication factor, should be between 0 and 1 (default is 0.5).
        """
        if self.step_dq < 0:
            self.step_dq = min(factor * self.step_dq, self.min_dq)
        else:
            self.step_dq = max(factor * self.step_dq, self.min_dq)

    def increase_rate_step(self, factor=1.5):
        """
        React to a deviation from the safe operating envelope.

        This function applies a multiplication factor greater than 1 to reduce the current rate step.
        The rate step is limited to a pre-defined maximum value.

        Parameters
        --------
        factor: float
            Multiplication factor, should be greater than 1 (default is 1.5).
        """
        if self.step_dq < 0:
            self.step_dq = max(factor * self.step_dq, self.max_dq)
        else:
            self.step_dq = min(factor * self.step_dq, self.max_dq)

    def react_to_deviation(self, t_deviation):
        """
        React to a deviation from the safe operating envelope.

        This function applies the following actions when a deviation from the safe operating envelope is detected:
            1. the injection rate is set to the rate from the previous step;
            2. the step start time is set to the time when the deviation is detected;
            3. the step end time is set to the new start time plus the step duration.

        Parameters
        --------
        t_deviation: float
            The time at which the deviation is detected, in seconds (relative to the simulation start time).
        """
        self.return_to_previous_injection_step = True
        self.q_inj = self.q_inj_previous_step
        self.step_start_time = t_deviation
        self.step_end_time = self.step_start_time + self.step_dt

    def update_parameters_for_next_step(self):
        """
        Update controller parameters for the next injection rate step.

        This function applies the following actions when a rate step is completed:
            1. the previous step injection rate is updated;
            2. the step index is incremented by 1;
            3. the step start and end times are updated for the next step;
            4. if the controller returned to the previous step as a result of deviation, the rate step and the
               injection rate are both reduced; otherwise (normal condition) the injection rate is increased.
        """
        self.q_inj_previous_step = self.q_inj
        self.step_index += 1
        self.step_start_time = self.step_end_time
        self.step_end_time = self.step_start_time + self.step_dt

        if self.return_to_previous_injection_step:
            self.return_to_previous_injection_step = False
            self.reduce_rate_step()
            self.q_inj -= self.step_dq
        else:
            self.q_inj += self.step_dq



import numpy as np
import pandas as pd
from scipy.interpolate import PchipInterpolator as interp_func


class PressureDeviationDetector:
    """
    Create PressureDeviationDetector class for monitoring of induced fractures and fracture opening in a step rate test.

    The class provides methods for storing pressure and rate data from each transient in the step rate test,
    calculation of Bourdet derivative, constructing a safe operating envelope based on the pressure transient history,
    and monitoring for deviations from the safe operating envelope.

    Attributes
    --------
    t_start_all_steps: list
        List of start times for each transient in hours, relative to the simulation start time.
    q_all_steps: list
        List of rates for each transient in m^3/s, to be used for superposition calculations.
    q_all_steps_pq: list
        List of rates for each transient in m^3/s, to be used for p-q plot.
    p_all_steps_pq: list
        List of final pressures for each transient in bars, to be used for p-q plot.
    t_all_steps_loglog_plot: list
        List of time arrays for each transient in hours, relative to the step start time.
    dp_all_steps: list
        List of normalized pressure transient arrays for each step, in bars.
    dp_der_all_steps: list
        List of pressure derivative arrays for each transient, in bars.
    dp_der_ref: np.array
        1d array of pressure derivative for the reference transient, in bars.
    dp_der_lower_bound: np.array
        1d array of pressure derivative lower bound, in bars.
    dp_der_upper_bound: np.array
        1d array of pressure derivative upper bound, in bars.
    dp_der_t_ref:  np.array
        1d array of time values for the reference transient, in hours.
    dp_der_t_soe: np.array
        1d array of time values for the safe operating envelope, in hours.
    q_ref: float
        Reference rate in m^3/s used for normalization of pressure transients. Sign convention is negative for injection.
    max_number_of_steps_for_reference_transient: int
        Maximum number of steps in the step rate test used for computing the reference transient derivative.
    detection_interval_start: float
        Time interval relative to the start of a transient when detection of deviations should start, in hours.
    detection_margin: float
        Log-scale margin used for deviation detection, relative to the reference transient.
    deviation_check_period: float
        Time period for checking deviations in hours.
    last_deviation_check_time: float
        Time relative to the start of a transient when the last deviation check occurred, in hours.
    weight_for_averaging: float
        Weight used for averaging the pressure derivative for the latest transient with the reference derivative.
    q_previous_step: float
        Rate at the previous step in m^3/s.
    deviation_detected: boolean
        Flag indicating if a deviation from the safe operating envelope has been detected.
    superposition_step_index: int
        The current step index for the superposition time calculation.

    Methods
    --------
    update_pressure_rate_history():
        Store pressure and rate data from the current transient in the step rate test.
    calculate_bourdet_derivative():
        Calculate the Bourdet pressure derivative for the current transient in the step rate test.
    update_safe_operating_envelope():
        Update the safe operating envelope based on the reference transient or from user-defined input.
    check_for_deviations():
        Check for pressure derivative deviations from the safe operating envelope.
    export_pressure_and_derivative_to_excel():
        Export pressure and derivative data to Excel format.
    """

    def __init__(
            self,
            q_ref=-50.0 / 3600,
            max_number_of_steps_for_reference_transient=3,
            detection_interval_start=1.0,
            detection_margin=0.2
    ):
        self.t_start_all_steps = []
        self.q_all_steps = []
        self.q_all_steps_pq = []
        self.p_all_steps_pq = []
        self.t_all_steps_loglog_plot = []
        self.dp_all_steps = []
        self.dp_der_all_steps = []
        self.dp_der_ref = []
        self.dp_der_lower_bound = []
        self.dp_der_upper_bound = []
        self.dp_der_t_ref = []
        self.dp_der_t_soe = []
        self.q_ref = q_ref
        self.max_number_of_steps_for_reference_transient = max_number_of_steps_for_reference_transient
        self.detection_interval_start = detection_interval_start
        self.detection_margin = detection_margin
        self.deviation_check_period = 0.1
        self.last_deviation_check_time = 0
        self.weight_for_averaging = 0.5
        self.q_previous_step = 0
        self.deviation_detected = False
        self.superposition_step_index = 0

    def update_pressure_rate_history(self, t, q, p_bh):
        """
        Store pressure and rate data from the current transient in the step rate test.

        This function updates the lists containing the rate and final well bottom-hole pressure for each transient in
        the step rate test, as long as no deviation in the pressure derivative was found in that transient.
        The time at the start of the step is also recorded, and the superposition step index is incremented by 1.
        This function should be called only when a step is completed.

        Parameters
        --------
        t: np.array
            1d array of time values for the current step, in hours.
        q: float
            Rate in m^3/s for the current step.
        p_bh: np.array
            1d array of well bottom-hole pressure for the current step, in bars.
        """
        if self.deviation_detected == False:
            self.t_start_all_steps.append(t[0])
            self.q_all_steps_pq.append(q)
            self.p_all_steps_pq.append(p_bh[-1])

        self.q_previous_step = q
        self.q_all_steps.append(q)
        self.superposition_step_index += 1

    def calculate_bourdet_derivative(self, superposition_step_index, t, q, p_bh, store_results=False):
        """
        Calculate the Bourdet pressure derivative for the current transient in the step rate test.

        This function first computes the superposition time based on the logarithmic time and the entire rate history.
        It then finds the pressure difference dp, normalized by rate.

        The local derivative is computed using a weighted average of pressure differences over the superposition time:
            dp_der = ((dp1 / dx1 * dx2 + dp2 / dx2 * dx1) / (dx1 + dx2))
        where:
        - dx1, dx2 are superposition time differences for the current and next interval.
        - dp1, dp2 are pressure differences for the current and next interval.

        The array of Bourdet derivatives is constructed for the entire transient, and if the store_results parameter
        is True, the array is stored together with the pressure difference and time points used in the calculation.

        Parameters
        --------
        superposition_step_index: int
            The current step index for the superposition time calculation.
        t: np.array
            1d array of time values for the current step, in hours.
        q: float
            Rate in m^3/s for the current step.
        p_bh: np.array
            1d array of well bottom-hole pressure for the current step, in bars.
        store_results: boolean
            Flag indicating whether to store the calculation results (default is False).

        Returns
        --------
        dp_der: np.array
            1d array of Bourdet pressure derivative in bars.
        """
        q_old = q
        dp_der_vals = []
        t_vals = []
        t_s = np.zeros(t.shape)
        min_dx = 1.0e-3  # minimum difference between two consecutive superposition times

        # Compute normalized pressure for the current step
        if superposition_step_index == 1:
            dq = q
        elif len(self.q_all_steps) < superposition_step_index:
            dq = q - self.q_all_steps[-1]
        else:
            dq = q - self.q_all_steps[-2]

        dp = np.abs((p_bh - p_bh[0]) / dq * self.q_ref).reshape(-1)

        # Compute superposition time
        if dq > 0:
            if len(self.q_all_steps) < superposition_step_index:
                q = self.q_all_steps[-1]
            else:
                q = self.q_all_steps[-2]

        for step_index in range(1, superposition_step_index + 1):
            if step_index == 1:
                t_s = np.add(t_s, (self.q_all_steps[step_index - 1]) / q * np.log10(t, where=t > 0))
            elif step_index == superposition_step_index:
                t_s = np.add(t_s,
                             (q_old - self.q_all_steps[step_index - 2]) / q * np.log10(t - t[0], where=t - t[0] > 0))
            else:
                t_s = np.add(t_s, (self.q_all_steps[step_index - 1] - self.q_all_steps[step_index - 2]) / q * np.log10(
                    t - self.t_start_all_steps[step_index - 1], where=t - self.t_start_all_steps[step_index - 1] > 0))

        # Compute derivative
        for i in range(1, len(p_bh) - 2):
            dt = t - t[0]
            # check if the first superposition time interval is longer than 0.1 log time units
            j = np.where(dt > dt[i] * np.exp(0.1))[0]
            if len(j) > 0:
                j = j[0]
                dp1 = dp[j] - dp[i]
                dx1 = t_s[j] - t_s[i]
            else:
                break

            # check if the second superposition time interval is longer than 0.1 log time units
            k = np.where(dt > dt[j] * np.exp(0.1))[0]
            if len(k) > 0:
                k = k[0]
                dp2 = dp[k] - dp[j]
                dx2 = t_s[k] - t_s[j]
            else:
                break

            if np.abs(dx1) > min_dx and np.abs(dx2) > min_dx:
                dp_der = (dp1 / dx1 * dx2 + dp2 / dx2 * dx1) / (dx1 + dx2)
            else:
                break

            dp_der = dq / q * dp_der / np.log(10)

            if not (np.isnan(dp_der)) and not ((t[j] - t[0]) in t_vals):
                t_vals.append(t[j] - t[0])
                dp_der_vals.append(dp_der)

        # From the calculated time points interpolate the pressure derivative for the entire time array
        if len(t_vals) > 0:
            dp_der_func = interp_func(t_vals, dp_der_vals, extrapolate=False)
            dp_der = dp_der_func(t - t[0])
        else:
            dp_der = np.ones((len(dp))) * np.nan

        # If no deviations have been detected for the current transient, store the time, pressure difference and derivative array
        if self.deviation_detected == False and store_results:
            self.t_all_steps_loglog_plot.append(t - t[0])
            self.dp_all_steps.append(dp)
            self.dp_der_all_steps.append(dp_der)

        return dp_der

    def update_safe_operating_envelope(self, dp_der_t_soe=None, dp_der_lower_bound=None, dp_der_upper_bound=None):
        """
        Update the safe operating envelope based on the reference transient or from user-defined input.

        This function first updates the reference transient pressure derivative based on the pressure transient history.
        The first transient in the step rate test is used as the initial reference transient, together with the
        corresponding pressure derivative and time points. The derivative is processed to remove NaN values.
        For subsequent steps, the reference derivative is recomputed by averaging the pressure derivative values
        for that step and the previous reference derivative. This is only done if no deviation from the safe operating
        envelope is found for that particular transient, and for a limited number of steps.
        Once the reference derivative has been updated, the detection margin is added symmetrically around it to obtain
        the lower and upper bounds of the safe operating envelope. The function allows the option to provide these
        bounds directly by the user (with the corresponding time points), in which case they will be used instead of
        the ones computed from the reference derivative.

        Parameters
        --------
        dp_der_t_soe : np.array
            1d array of time values for the safe operating envelope, in hours.
        dp_der_lower_bound : np.array
            1d array of pressure derivative lower bound, in bars.
        dp_der_upper_bound : np.array
            1d array of pressure derivative upper bound, in bars.
        """
        if len(self.dp_der_all_steps) == 0:
            raise AttributeError(
                'No pressure derivatives have been computed. Call calculate_bourdet_derivative() first.')
        if len(self.dp_der_all_steps) == 1:
            self.dp_der_ref = self.dp_der_all_steps[0]  # reference transient pressure derivative
            self.dp_der_t_ref = self.t_all_steps_loglog_plot[0]  # time array for reference transient
            idx_nan = np.isnan(self.dp_der_ref)
            self.dp_der_ref = self.dp_der_ref[~idx_nan]
            self.dp_der_t_ref = self.dp_der_t_ref[~idx_nan]
        elif len(
                self.dp_der_all_steps) <= self.max_number_of_steps_for_reference_transient and self.deviation_detected == False:
            idx_nan = np.isnan(self.dp_der_all_steps[-1])
            dp_der_func = interp_func(self.t_all_steps_loglog_plot[-1][~idx_nan],
                                      self.dp_der_all_steps[-1][~idx_nan],
                                      extrapolate=False)
            for k in range(len(self.dp_der_ref)):
                if not (np.isnan(dp_der_func(self.dp_der_t_ref[k]))):
                    self.dp_der_ref[k] = self.dp_der_ref[k] * (1 - self.weight_for_averaging) + dp_der_func(
                        self.dp_der_t_ref[k]) * self.weight_for_averaging

        if (dp_der_t_soe is None) or (dp_der_lower_bound is None) or (dp_der_upper_bound is None):
            if len(self.dp_der_ref) > 0:
                self.dp_der_lower_bound = np.exp(-self.detection_margin) * self.dp_der_ref
                self.dp_der_upper_bound = np.exp(self.detection_margin) * self.dp_der_ref
                self.dp_der_t_soe = self.dp_der_t_ref
        else:
            self.dp_der_lower_bound = dp_der_lower_bound
            self.dp_der_upper_bound = dp_der_upper_bound
            self.dp_der_t_soe = dp_der_t_soe

        self.deviation_detected = False  # if a deviation was detected previously, we set the flag back to False here

    def check_for_deviations(self, step_index, t, p_bh, q):
        """
        Check for pressure derivative deviations from the safe operating envelope.

        This function computes the Bourdet derivative array for the current pressure transient and compares each
        derivative after the detection start time to the upper and lower bounds of the safe operating envelope. If the
        derivative is outside those bounds for any value in the interval investigated, the deviation_detected flag is
        set to True. The pressure and derivative up to the time when the deviation is detected are stored in the lists
        containing the pressure transient data.

        Parameters
        --------
        step_index: int
            The current step index in the step rate test.
        t: np.array
            1d array of time values for the current step, in hours.
        q: float
            Rate in m^3/s for the current step.
        p_bh: np.array
            1d array of well bottom-hole pressure for the current step, in bars.

        Returns
        --------
        deviation_detected: boolean
            Flag indicating if a deviation from the safe operating envelope has been detected.
        """
        if step_index > 1:
            if len(self.dp_der_t_soe) > 1:
                # set up interpolation functions for the safe operating envelope bounds, since they may be evaluated at
                # different time points than the current pressure transient
                dp_der_lower_bound_func = interp_func(self.dp_der_t_soe, self.dp_der_lower_bound, extrapolate=False)
                dp_der_upper_bound_func = interp_func(self.dp_der_t_soe, self.dp_der_upper_bound, extrapolate=False)

                # continue only if the time elapsed since the previous check is longer than the detection check period
                if len(t) > 0 and t[-1] - self.deviation_check_period > self.last_deviation_check_time:
                    t0 = t[0]
                    self.last_deviation_check_time = t[-1]
                    t_loglog = t - t0

                    # check if the time interval is longer than the detection start time
                    if t_loglog[-1] > self.detection_interval_start:
                        dp_der = self.calculate_bourdet_derivative(self.superposition_step_index + 1, t_loglog + t0, q,
                                                                   p_bh)
                        j = np.where(t_loglog > self.detection_interval_start)[0][0]  # interval start index
                        k = np.where(dp_der > 0)[0][-1]  # find index of last valid pressure derivative
                        dq = q - self.q_previous_step
                        dp = np.abs((p_bh - p_bh[0]) / dq * self.q_ref)  # normalized pressure difference
                        index_interval_end = min(k, len(t_loglog))
                        index_interval_start = j

                        for i in range(index_interval_start, index_interval_end):
                            if (dp_der[i] < dp_der_lower_bound_func(t_loglog[i]) or dp_der[i] > dp_der_upper_bound_func(
                                    t_loglog[i])):
                                print(f'Deviation from safe operating envelope detected at injection step {step_index},'
                                      f' pressure={np.round(p_bh[i], 2)} bar')

                                # Update pressure and rate data from the current transient
                                self.t_start_all_steps.append(t[0])
                                self.q_previous_step = q
                                self.q_all_steps.append(q)
                                self.q_all_steps_pq.append(q)
                                self.p_all_steps_pq.append(p_bh[-1])
                                self.t_all_steps_loglog_plot.append(t - t[0])
                                self.dp_all_steps.append(dp)
                                self.dp_der_all_steps.append(dp_der)
                                self.superposition_step_index += 1
                                self.t_start_all_steps.append(t[-1])
                                self.deviation_detected = True

                                return self.deviation_detected
            else:
                raise AttributeError(
                    'No safe operating envelope has been defined. Call update_safe_operating_envelope() first.')

        return self.deviation_detected

    def export_pressure_and_derivative_to_excel(self, name):
        """
        Export pressure and derivative data to Excel format.

        This function exports to an Excel file the normalized pressure and derivative together with the time points
        used in the derivative calculation for each transient in the step rate test.

        Parameters
        --------
        name: str
            File name without the format extension.
        """
        writer = pd.ExcelWriter(f'{name}.xlsx', engine='openpyxl')

        for i in range(len(self.t_all_steps_loglog_plot)):
            df = pd.DataFrame({"Elapsed Time (hr)": self.t_all_steps_loglog_plot[i].reshape(-1),
                               "Pressure (bar)": self.dp_all_steps[i].reshape(-1),
                               "Pressure derivative (bar)": self.dp_der_all_steps[i].reshape(-1)}).dropna()
            sheet_name = 'Step_' + str(i + 1)
            df.to_excel(writer, sheet_name=f'{sheet_name}', index=False)
        writer.close()



import numpy as np
import pandas as pd


class ReservoirSimulator:
    """
    Create ReservoirSimulator class for transient, single-phase reservoir model with pressure-dependent permeability.

    The class provides methods for solving the radial diffusivity equation for both constant and pressure-dependent
    permeability.

    Attributes
    --------
    k: float
        The reservoir matrix permeability in m^2.
    phi: float
        The reservoir porosity, defined as a fraction.
    mu: float
        The fluid viscosity in Pa.s.
    c: float
        The combined compressibility of the rock and fluid in Pa^-1.
    b: float
        The formation volume factor in m^3/stm^3.
    alpha: float
        The diffusivity coefficient in m^2/s.
    p_res: float
        The far-field reservoir pressure in Pa.
    ri: float
        The wellbore radius in m.
    ro: float
        The radial distance to the reservoir outer boundary in m.
    h: float
        The reservoir thickness in m.
    s: float
        The dimensionless skin factor.
    q: float
        The rate in m^3/s. Sign convention is negative for injection and positive for production.
    c_wb: float
        The wellbore storage constant in m^3/Pa.
    p_frac: float
        The fracture opening pressure in Pa.
    xf: float
        The fracture half-length in m.
    t_max: float
        The maximum simulation time in seconds.
    dt: float
        The time step of the solver outer loop in seconds.
    dt_i: float
        The time step of the solver inner loop in seconds.
    t: np.array
        1d array of simulation time values in seconds, where a constant time step equal to dt is used.
    nt: int
        The size of the simulation time array.
    nx: int
        The size of the simulation spatial grid.
    use_uniform_grid: boolean
        If this is True, a grid with uniform spacing is used, otherwise a logarithmic spacing is used.
    use_pressure_dependent_permeability: boolean
        If this is True, the simulator uses a pressure-dependent permeability function, otherwise uses constant permeability.
    use_rate_limitation: boolean
        If this is True, a low pass filter is applied on the rate command to get smoother numerical solution during rate changes.
    rate_low_pass_filter_constant: float
        The rate low-pass filter time constant in seconds.
    outer_boundary_condition: int
        The boundary condition type at the outer boundary: 0 = no-flow (closed circle), 1 = constant pressure (default).
    x: np.array
        1d array of the simulation spatial grid in m.
    dx: float
        Grid spacing in m. For logarithmic grid, this is the spacing near the wellbore.
    idx_frac: int
        Index of fracture half-length in the spatial grid.
    p: np.array
        2d array (space and time) of the simulated reservoir pressure in Pa.
    p_bh: np.array
        1d array of the simulated well bottom-hole pressure in Pa.
    q_hist: np.array
        1d array of the rate history in m^3/s. Sign convention is negative for injection and positive for production.

    Methods
    --------
    permeability_function():
        Compute the pressure-dependent permeability.
    solve_tdma():
        Solve a tridiagonal system of equations using the TDMA algorithm.
    step():
        Solve the radial diffusivity equation for one time step.
    solve_one_step_constant_permeability():
        Solve the radial diffusivity equation with constant permeability for one time step.
    solve_one_step_pressure_dependent_permeability():
        Solve the radial diffusivity equation with pressure-dependent permeability for one time step.
    export_pressure_and_rate_to_csv():
        Export simulated well bottom-hole pressure and rate history to csv format.
    export_pressure_and_rate_to_excel():
        Export simulated well bottom-hole pressure and rate history to Excel format.
    """

    def __init__(self,
                 dt=10,
                 t_max=150 * 3600,
                 k = 100 * 1e-15,
                 phi = 0.3,
                 mu = 0.00032,
                 c = 8.6e-10,
                 b = 1.03,
                 p_res = 250e5,
                 ri = 0.1,
                 ro = 10000,
                 h = 15,
                 s = 0,
                 c_wb = 1.5e-7,
                 p_frac = 380e5,
                 xf = 5,
                 use_pressure_dependent_permeability=True
                 ):
        ## Set physical parameters
        self.k = k
        self.phi = phi
        self.mu = mu
        self.c = c
        self.b = b
        self.alpha = self.k / (self.mu * self.phi * self.c)
        self.p_res = p_res
        self.ri = ri
        self.ro = ro
        self.h = h
        self.s = s
        self.q = 0
        self.c_wb = c_wb
        self.p_frac = p_frac
        self.xf = xf

        if ri >= ro:
            raise ValueError('The wellbore radius needs to be smaller than the reservoir outer boundary radius.')

        if xf >= ro:
            raise ValueError('The fracture half-length needs to be smaller than the reservoir outer boundary radius.')

        ## Set simulation parameters
        self.t_max = t_max
        self.dt = dt
        self.dt_i = 1.0
        self.t = np.arange(0, self.t_max + self.dt, self.dt)
        self.nt = len(self.t)
        self.nx = 100

        self.use_uniform_grid = False
        self.use_pressure_dependent_permeability = use_pressure_dependent_permeability
        self.use_rate_limitation = False
        self.rate_low_pass_filter_constant = self.dt / 20
        self.outer_boundary_condition = 1

        if self.use_uniform_grid:
            self.x = np.transpose(np.linspace(self.ri, self.ro, self.nx))
            self.dx = (self.ro - self.ri) / (self.nx - 1)  # constant grid spacing
        else:
            self.x = np.transpose(np.logspace(np.log10(self.ri), np.log10(self.ro), self.nx))
            self.dx = self.x[1] - self.x[0]  # grid spacing near the wellbore

        self.idx_frac = next(
            x for x, val in enumerate(self.x) if val > self.xf)

        self.p = np.full([self.nx, self.nt], np.nan)
        self.p_bh = np.full(self.nt, np.nan)
        self.q_hist = np.full(self.nt, np.nan)

        ## Initialize states
        self.p[:, 0] = np.ones((self.nx, 1)).reshape(-1) * self.p_res
        self.p_bh[0] = self.p_res
        self.q_hist[0] = 0

    def permeability_function(self, p):
        """
        Compute the pressure-dependent permeability.

        This function computes the reservoir permeability as a function of reservoir pressure, as a proxy-model
        for fracture permeability. It is defined as an exponential function with saturation, such that where the
        reservoir pressure is below a threshold, representing the fracture opening pressure, the permeability is equal to
        the matrix permeability. When it is above the fracture opening pressure, it increases exponentially, until some
        maximum ratio of fracture permeability to matrix permeability, which is defined inside this function.
        The output of this function is an array of permeability values with the same size as the pressure array.

        Parameters
        --------
        p: np.array
            1d array of the reservoir pressure for one simulation step in Pa.

        Returns
        --------
        kp: np.array
            1d array of reservoir permeability in m^2.
        """
        k_ratio_min = 1.0  # minimum permeability ratio
        k_ratio_max = 4.0  # maximum permeability ratio
        a = 0.008  # the exponent of the exponential function
        kp = self.k * np.maximum(np.minimum(np.exp(a * (p - self.p_frac) / 1e5), k_ratio_max),
                                 k_ratio_min)

        return kp

    def solve_tdma(self, a, b, c, d):
        """
        Solve a tridiagonal system of equations using the TDMA algorithm.

        This function uses the tridiagonal matrix algorithm (TDMA) to solve the system of linear equations:
            A*x=d

        where:
        - x is the solution vector.
        - A is the tridiagonal matrix split into the vectors a (lower diagonal), b (main diagonal) and c (upper diagonal).
        - d is the right-hand side vector.
        (http://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm)

        Parameters
        --------
        a: np.array
            1d array containing the elements of the matrix lower diagonal.
        b: np.array
            1d array containing the elements of the matrix main diagonal.
        c: np.array
            1d array containing the elements of the matrix upper diagonal.
        d: np.array
            1d array of the right-hand side elements.

        Returns
        --------
        x: np.array
            1d array containing the solution of the system of linear equations.
        """
        n = len(d)  # number of equations to solve
        a_prime, b_prime, c_prime, d_prime = map(np.array, (a, b, c, d))
        for i in range(1, n):  # forward elimination
            m = a_prime[i] / b_prime[i - 1]
            b_prime[i] = b_prime[i] - m * c_prime[i - 1]
            d_prime[i] = d_prime[i] - m * d_prime[i - 1]

        x = b_prime.copy()
        x[-1] = d_prime[-1] / b_prime[-1]

        for j in range(n - 2, -1, -1):  # backward substitution
            x[j] = (d_prime[j] - c_prime[j] * x[j + 1]) / b_prime[j]

        return x

    def step(self, p0, q_c):
        """
        Solve the radial diffusivity equation for one step.

        This function solves the radial diffusivity equation for one time step. It takes the rate command as input and
        passes it to the solver. If the use_rate_limitation parameter is True, it applies a low-pass filter on the rate
        command before passing it to the solver. The function chooses which solver to use based on the value of the
        use_pressure_dependent_permeability parameter, and proceeds into the corresponding solver.

        Parameters
        --------
        p0: np.array
            1d array of the reservoir pressure from the previous time step in Pa.
        q_c: float
            The rate command in m^3/s. Sign convention is negative for injection and positive for production.

        """
        if self.use_rate_limitation:
            self.q = self.q * (1 - self.rate_low_pass_filter_constant) + self.rate_low_pass_filter_constant * q_c
        else:
            self.q = q_c

        if self.use_pressure_dependent_permeability:
            return self.solve_one_step_pressure_dependent_permeability(p0)
        else:
            return self.solve_one_step_constant_permeability(p0)

    def solve_one_step_constant_permeability(self, p0):
        """
        Solve the radial diffusivity equation with constant permeability for one time step.

        This function solves the radial diffusivity equation for one time step for constant reservoir permeability.
        The diffusivity equation is discretized using the Crank-Nicolson finite difference method.
        (https://en.wikipedia.org/wiki/Crank-Nicolson_method)

        This formulation results in a tridiagonal system of equations which is solved with the solve_tdma() method.

        Parameters
        --------
        p0: np.array
            1d array of the reservoir pressure from the previous time step in Pa.

        Returns
        --------
        p: np.array
            1d array of the reservoir pressure for the current time step in Pa.
        p_bh: float
            Well bottom-hole pressure for the current time step in Pa.
        """
        # Parse input
        p = p0

        # Define internal solver time step
        Nt = max(int(np.floor(self.dt / self.dt_i)), 1)
        dt = self.dt / Nt

        a_matrix = np.zeros((self.nx, self.nx))
        # Assemble matrix from finite difference discretization
        for i in range(1, self.nx - 1):
            a_matrix[i, i - 1] = 0.5 * self.alpha / self.x[i] / (self.x[i + 1] - self.x[i - 1]) - self.alpha / (
                    self.x[i] - self.x[i - 1]) / (self.x[i + 1] - self.x[i - 1])
            a_matrix[i, i] = 1.0 / dt + self.alpha / (self.x[i + 1] - self.x[i]) / (self.x[i] - self.x[i - 1])
            a_matrix[i, i + 1] = - 0.5 * self.alpha / self.x[i] / (self.x[i + 1] - self.x[i - 1]) - self.alpha / (
                    self.x[i + 1] - self.x[i]) / (self.x[i + 1] - self.x[i - 1])
        a_matrix[self.nx - 1, self.nx - 1] = 1

        # account for boundary condition type in the tridiagonal matrix coefficients
        if self.outer_boundary_condition == 0:
            a_matrix[self.nx - 1, self.nx - 2] = - 1

        # account for wellbore storage in the tridiagonal matrix coefficients
        a_matrix[0, 0] = - 1 - self.dx * self.mu * self.c_wb * (1 - self.s * self.ri / self.dx) / (
                2 * np.pi * self.ri * self.h * self.k * dt)
        a_matrix[0, 1] = 1 - self.mu * self.c_wb * self.s / (2 * np.pi * self.h * self.k * dt)

        # Solver inner loop
        for n in range(Nt):
            # assemble right-hand side terms, accounting for wellbore storage and boundary condition type
            b = np.zeros((self.nx, 1)).reshape(-1)
            b[0] = self.dx * self.mu / (2 * np.pi * self.ri * self.h * self.k * dt) * (
                    self.q * self.b * dt - self.c_wb * (p[0] + self.s * self.ri / self.dx * (p[1] - p[0])))

            if self.outer_boundary_condition == 1:
                b[self.nx - 1] = self.p_res

            for i in range(1, self.nx - 1):
                b[i] = (1.0 / dt - self.alpha / (self.x[i + 1] - self.x[i]) / (self.x[i] - self.x[i - 1])) * p[i] - \
                       a_matrix[
                           i, i + 1] * p[i + 1] - a_matrix[i, i - 1] * p[i - 1]

            # collect main, lower and upper diagonal from the system matrix
            a_p = np.array([a_matrix[i, i] for i in range(self.nx)])  # main diagonal
            a_w = [a_matrix[i, i - 1] for i in range(1, self.nx)]  # lower diagonal
            a_e = [a_matrix[i, i + 1] for i in range(0, self.nx - 1)]  # upper diagonal
            a_w = np.concatenate([[0], a_w])
            a_e = np.concatenate([a_e, [0]])
            p = self.solve_tdma(a_w, a_p, a_e, b)

        ## Parse output
        p = p.reshape(-1)
        p_bh = p[0] - self.s * b[0] * self.ri / self.dx

        return p, p_bh

    def solve_one_step_pressure_dependent_permeability(self, p0):
        """
        Solve the radial diffusivity equation with pressure-dependent permeability for one time step.

        This function solves the radial diffusivity equation for one time step for pressure-dependent reservoir
        permeability. The diffusivity equation is solved using the Crank-Nicolson finite difference method.
        (https://en.wikipedia.org/wiki/Crank-Nicolson_method)

        The permeability for grid cells in the fractured zone defined by the fracture half-length (xf) parameter is
        updated for each internal solver time step, while for the remaining grid cells, the matrix permeability is used.
        This formulation results in a tridiagonal system of equations which is solved with the solve_tdma() method.

        Parameters
        --------
        p0: np.array
            1d array of the reservoir pressure from the previous time step in Pa.

        Returns
        --------
        p: np.array
            1d array of the reservoir pressure for the current time step in Pa.
        p_bh: float
            Well bottom-hole pressure for the current time step in Pa.
        """
        # Parse input
        p = p0

        # Define internal solver time step
        Nt = max(int(np.floor(self.dt / self.dt_i)), 1)
        dt = self.dt / Nt

        a_matrix = np.zeros((self.nx, self.nx))
        kp = np.ones(p.shape) * self.k

        # Solver inner loop
        for n in range(Nt):
            # update permeability for grid cells in the fractured zone
            kp[:self.idx_frac] = self.permeability_function(p[:self.idx_frac])

            # assemble matrix from finite difference discretization
            for i in range(1, self.nx - 1):
                a_matrix[i, i - 1] = 0.5 * kp[i] / (self.mu * self.phi * self.c) / self.x[i] / (
                        self.x[i + 1] - self.x[i - 1]) + 0.5 * (
                                             kp[i + 1] - kp[i - 1]) / (self.mu * self.phi * self.c) / (
                                             self.x[i + 1] - self.x[i - 1]) / (
                                             self.x[i + 1] - self.x[i - 1]) - kp[i] / (self.mu * self.phi * self.c) / (
                                             self.x[i] - self.x[i - 1]) / (self.x[i + 1] - self.x[i - 1])
                a_matrix[i, i] = 1.0 / dt + kp[i] / (self.mu * self.phi * self.c) / (self.x[i + 1] - self.x[i]) / (
                        self.x[i] - self.x[i - 1])
                a_matrix[i, i + 1] = - 0.5 * kp[i] / (self.mu * self.phi * self.c) / self.x[i] / (
                        self.x[i + 1] - self.x[i - 1]) - 0.5 * (kp[i + 1] - kp[i - 1]) / (
                                             self.mu * self.phi * self.c) / (
                                             self.x[i + 1] - self.x[i - 1]) / (self.x[i + 1] - self.x[i - 1]) - kp[
                                         i] / (
                                             self.mu * self.phi * self.c) / (self.x[i + 1] - self.x[i]) / (
                                             self.x[i + 1] - self.x[i - 1])
            a_matrix[self.nx - 1, self.nx - 1] = 1

            # account for boundary condition type in the tridiagonal matrix coefficients
            if self.outer_boundary_condition == 0:
                a_matrix[self.nx - 1, self.nx - 2] = - 1

            # account for wellbore storage in the tridiagonal matrix coefficients
            a_matrix[0, 0] = - 1 - self.dx * self.mu * self.c_wb * (1 - self.s * self.ri / self.dx) / (
                    2 * np.pi * self.ri * self.h * kp[0] * dt)
            a_matrix[0, 1] = 1 - self.mu * self.c_wb * self.s / (2 * np.pi * self.h * kp[0] * dt)

            # assemble right-hand side terms, accounting for wellbore storage and boundary condition type
            b = np.zeros((self.nx, 1)).reshape(-1)
            b[0] = self.dx * self.mu / (2 * np.pi * self.ri * self.h * kp[0] * dt) * (
                    self.q * self.b * dt - self.c_wb * (
                    p[0] + self.s * self.ri / self.dx * (p[1] - p[0])))

            if self.outer_boundary_condition == 1:
                b[self.nx - 1] = self.p_res

            for i in range(1, self.nx - 1):
                b[i] = (1.0 / dt - kp[i] / (self.mu * self.phi * self.c) / (self.x[i + 1] - self.x[i]) / (
                        self.x[i] - self.x[i - 1])) * p[i] - a_matrix[i, i + 1] * p[i + 1] - a_matrix[i, i - 1] * p[
                           i - 1]

            # collect main, lower and upper diagonal from the system matrix
            a_p = np.array([a_matrix[i, i] for i in range(self.nx)])  # main diagonal
            a_w = [a_matrix[i, i - 1] for i in range(1, self.nx)]  # lower diagonal
            a_e = [a_matrix[i, i + 1] for i in range(0, self.nx - 1)]  # upper diagonal
            a_w = np.concatenate([[0], a_w])
            a_e = np.concatenate([a_e, [0]])
            p = self.solve_tdma(a_w, a_p, a_e, b)

        ## Parse output
        p = p.reshape(-1)
        p_bh = p[0] - self.s * b[0] * self.ri / self.dx

        return p, p_bh

    def export_pressure_and_rate_to_csv(self, name):
        """
        Export simulated well bottom-hole pressure and rate history to csv format.

        This function exports to a csv file the simulated well bottom-hole pressure and the rate history for
        each simulation time step.

        Parameters
        --------
        name: str
            File name without the format extension.
        """
        df = pd.DataFrame(
            {"Time (hr)": np.array(self.t) / 3600, "Pressure (bar)": np.array(self.p_bh.reshape(-1)) / 1e5,
             "Rate (m3/D)": np.array(self.q_hist.reshape(-1)) * 3600 * 24}).dropna()
        df.to_csv(f'{name}.csv', index=False)

    def export_pressure_and_rate_to_excel(self, name):
        """
        Export simulated well bottom-hole pressure and rate history to Excel format.

        This function exports to an Excel file the simulated well bottom-hole pressure and the rate history for
        each simulation time step.

        Parameters
        --------
        name: str
            File name without the format extension.
        """
        df = pd.DataFrame(
            {"Time (hr)": np.array(self.t) / 3600, "Pressure (bar)": np.array(self.p_bh.reshape(-1)) / 1e5,
             "Rate (m3/D)": np.array(self.q_hist.reshape(-1)) * 3600 * 24}).dropna()

        writer = pd.ExcelWriter(f'{name}.xlsx', engine='openpyxl')
        df.to_excel(writer, index=False)
        writer.close()
